import torch
from einops import rearrange

MODEL = None
TOKENIZER = None
BERT_VOCAB_SIZE = 30522

def exists(val):
    return val is not None

def get_tokenizer():
    global TOKENIZER
    if not exists(TOKENIZER):
        TOKENIZER = torch.hub.load('huggingface/pytorch-transformers', 'tokenizer', 'bert-base-cased')
    return TOKENIZER

def get_bert():
    global MODEL
    if not exists(MODEL):
        MODEL = torch.hub.load('huggingface/pytorch-transformers', 'model', 'bert-base-cased')
    return MODEL

def tokenize(texts, add_special_tokens = True):
    if not isinstance(texts, (list, tuple)):
        texts = [texts]

    tokenizer = get_tokenizer()

    encoding = tokenizer.batch_encode_plus(
        texts,
        add_special_tokens = add_special_tokens,
        padding = True,
        return_attention_mask = True,
        return_tensors = 'pt'
    )

    token_ids = encoding.input_ids
    mask = encoding.attention_mask
    return token_ids, mask

def bert_embed(
    token_ids,
    mask = None,
    return_cls_repr = False,
    eps = 1e-8
):
    model = get_bert()

    outputs = model(
        input_ids = token_ids,
        attention_mask = None,
        output_hidden_states = True
    )

    hidden_state = outputs.hidden_states[-1]

    if return_cls_repr:
        return hidden_state[:, 0]               # return [cls] as representation

    if not exists(mask):
        return hidden_state.mean(dim = 1)

    mask = mask[:, 1:]                          # mean all tokens excluding [cls], accounting for length
    mask = rearrange(mask, 'b n -> b n 1')

    numer = (hidden_state[:, 1:] * mask).sum(dim = 1)
    denom = mask.sum(dim = 1)
    masked_mean =  numer / (denom + eps)
    return masked_mean



from functools import partial
import numpy as np
import torch
from torch.utils.data import Dataset
from contextlib import contextmanager

@contextmanager
def memmap(*args, **kwargs):
    pointer = np.memmap(*args, **kwargs)
    yield pointer
    del pointer

# dataset

class RETRODataset(Dataset):
    def __init__(
        self,
        *,
        num_chunks,
        chunk_size,
        num_sequences,
        num_neighbors,
        chunk_memmap_path,
        chunk_continuation_memmap_path,
        chunk_nn_memmap_path,
        seq_memmap_path,
        mask_memmap_path = None,
        seq_len = 2048
    ):
        super().__init__()
        self.num_sequences = num_sequences
        self.seq_num_chunks = seq_len // chunk_size
        shape = (num_chunks, chunk_size + 1)

        self.get_chunks = partial(memmap, chunk_memmap_path, dtype = np.int32, shape = shape)
        self.get_masks = partial(memmap, mask_memmap_path, dtype = np.bool, shape = shape)
        self.get_continuations = partial(memmap, chunk_continuation_memmap_path, dtype = np.int32, shape = (num_chunks,))
        self.get_knns = partial(memmap, chunk_nn_memmap_path, dtype = np.int32, shape = (num_chunks, num_neighbors))
        self.get_seqs = partial(memmap, seq_memmap_path, dtype = np.int32, shape = (num_sequences,))

    def __len__(self):
        return self.num_sequences

    def __getitem__(self, ind):
        with self.get_chunks() as chunks_memmap, self.get_masks() as masks_memmap, self.get_continuations() as continuations_memmap, self.get_knns() as knns_memmap, self.get_seqs() as seqs_memmap:

            begin_chunk_index = seqs_memmap[ind]
            chunk_range = slice(begin_chunk_index, (begin_chunk_index + self.seq_num_chunks))

            chunks = chunks_memmap[chunk_range]

            # excise the last token, except for last token of last chunk
            seq_tokens = np.concatenate((chunks[:, :-1].flatten(), chunks[-1, -1:]))

            # derive retrieved tokens
            knns = knns_memmap[chunk_range]
            continuations = continuations_memmap[knns]

            # get neighbor and continuation chunks
            knn_chunks = chunks_memmap[knns]
            knn_masks = masks_memmap[knns]

            continuation_chunks = chunks_memmap[continuations]
            continuation_masks = masks_memmap[continuations]

            # combine neighbors with continuations

            retrieved = np.concatenate((knn_chunks[..., :-1], continuation_chunks[..., :-1]), axis = -1)
            retrieved_masks = np.concatenate((knn_masks[..., :-1], continuation_masks[..., :-1]), axis = -1)

        return torch.from_numpy(seq_tokens).long(), torch.from_numpy(retrieved).long(), torch.from_numpy(retrieved_masks)
